package filegateway

import (
	"context"
	"path"
	"sort"
	"strings"
)

// FileObserver is a helper for emitting file names to a channel.
// It takes lists of files, optionnally filter with a pattern, and
// globally deduplicate the list
type FileObserver struct {
	pattern string
	prefix  string
	c       chan<- string
	ctx     context.Context

	files map[string]bool
}

// NewFileObserver creates a FileObserver
func NewFileObserver(ctx context.Context, pattern string, prefix string, c chan<- string) *FileObserver {
	if prefix != "" && !strings.HasSuffix(prefix, "/") {
		prefix += "/"
	}
	return &FileObserver{
		pattern,
		prefix,
		c,
		ctx,
		make(map[string]bool),
	}
}

// AddFiles add a list of files
func (fo *FileObserver) AddFiles(files ...string) error {
	sort.Strings(files)
	for _, filepath := range files {
		if fo.prefix != "" && strings.HasPrefix(filepath, fo.prefix) {
			filepath = filepath[len(fo.prefix):]
		}

		if fo.pattern != "" {
			m, err := path.Match(fo.pattern, filepath)
			if err != nil {
				return err
			}
			if !m {
				continue
			}
		}

		if _, ok := fo.files[filepath]; !ok {
			fo.files[filepath] = true
			select {
			case fo.c <- filepath:
			case <-fo.ctx.Done():
				return nil
			}
		}
	}
	return nil
}
