package stdlib

import (
	"fmt"
	"io"

	minio "github.com/minio/minio-go"
)

// S3Settings holds secrets and settings to connect to a s3 endpoint
type S3Settings struct {
	AccessKeyID     string
	SecretAccessKey string
	UseSSL          bool
}

var s3Endpoints = make(map[string]S3Settings)

// SetS3Endpoint set the settings for a given endpoint
func SetS3Endpoint(endpoint string, settings S3Settings) {
	s3Endpoints[endpoint] = settings
}

func s3Err(err error) error {
	return FileLocationErr{
		Type: "s3",
		Err:  err,
	}
}

// LoadS3Settings loads a endpoint settings
func LoadS3Settings(m map[string]interface{}) (S3Settings, error) {
	var settings S3Settings
	for key, value := range m {
		if key == "access-key" {
			if s, ok := value.(string); ok {
				settings.AccessKeyID = s
			} else {
				return settings, fmt.Errorf("invalid value for access-key. Expected a string, got: %s", value)
			}
		}
		if key == "secret-key" {
			if s, ok := value.(string); ok {
				settings.SecretAccessKey = s
			} else {
				return settings, fmt.Errorf("invalid value for secret-key. Expected a string, got: %s", value)
			}
		}
		if key == "ssl" {
			if b, ok := value.(bool); ok {
				settings.UseSSL = b
			} else {
				return settings, fmt.Errorf("invalid value for ssh. Expected a bool, got: %s", value)
			}
		}
	}
	return settings, nil
}

// LoadS3Endpoints loads all the s3 settings found in a map
func LoadS3Endpoints(m map[string]interface{}) error {
	for endpoint, value := range m {
		if entries, ok := value.(map[string]interface{}); ok {
			if settings, err := LoadS3Settings(entries); err == nil {
				SetS3Endpoint(endpoint, settings)
			} else {
				return s3Err(fmt.Errorf("invalid type for %s entry", endpoint))
			}
		}
	}
	return nil
}

// GetS3Endpoint returns the settings for the settings
func GetS3Endpoint(endpoint string) (*S3Settings, error) {
	if settings, ok := s3Endpoints[endpoint]; ok {
		return &settings, nil
	}
	return nil, s3Err(fmt.Errorf("no known settings for this s3 endpoint: %s", endpoint))
}

// FileLocationS3 implements FileLocationType for a s3 distant storage
type FileLocationS3 struct{}

// ReadBody returns a reader on the body
func (fls3 *FileLocationS3) ReadBody(location FileLocation) (io.ReadCloser, error) {
	endpoint, err := location.StringArg("provider")
	if err != nil {
		return nil, s3Err(err)
	}
	bucketName, err := location.StringArg("bucket-name")
	if err != nil {
		return nil, s3Err(err)
	}
	objectid, err := location.StringArg("objectid")
	if err != nil {
		return nil, s3Err(err)
	}
	settings, err := GetS3Endpoint(endpoint)
	if err != nil {
		return nil, s3Err(err)
	}
	minioClient, err := minio.New(
		endpoint, settings.AccessKeyID, settings.SecretAccessKey, settings.UseSSL)
	if err != nil {
		return nil, s3Err(err)
	}
	object, err := minioClient.GetObject(bucketName, objectid, minio.GetObjectOptions{})
	if err != nil {
		return nil, s3Err(err)
	}
	return WrapFileLocationReadCloser("s3", object), nil
}

func init() {
	RegisterFileLocationType("s3", &FileLocationS3{})
}
