package routing

import (
	"errors"
	"fmt"

	"orus.io/xbus/xbus-stdlib/service"
)

// ErrNoSuchKey is returned by Get if the key is not present in the store
var ErrNoSuchKey = errors.New("no such key")

// TableStore is a basic kv store interface
type TableStore interface {
	Set(key, value string) error
	Get(key string) (string, error)
	Delete(key string) error
	Close() error
}

// TableStoreFactory is a factory of TableStore
type TableStoreFactory func(prefix string, settings service.Settings) (TableStore, error)

var kvStoreTypes = map[string]TableStoreFactory{}

// RegisterTableStore registers a TableStore kind
func RegisterTableStore(name string, factory TableStoreFactory) error {
	if _, ok := kvStoreTypes[name]; ok {
		return fmt.Errorf("kvstore kind '%s' already exists", name)
	}
	kvStoreTypes[name] = factory
	return nil
}

// MustRegisterTableStore registers a TableStore kind and panics if it failed
func MustRegisterTableStore(name string, factory TableStoreFactory) {
	if err := RegisterTableStore(name, factory); err != nil {
		panic(err)
	}
}

// UnregisterTableStore unregisteres a TableStore kind
func UnregisterTableStore(name string) {
	delete(kvStoreTypes, name)
}

// NewTableStore creates a TableStore of the right kind
func NewTableStore(entryname string, settings service.Settings) (TableStore, error) {
	kind, err := settings.String(entryname + ".backend")
	if err != nil {
		return nil, err
	}

	if factory, ok := kvStoreTypes[kind]; ok {
		return factory(entryname+".", settings)
	}
	return nil, fmt.Errorf("no such TableStore backend: %s", kind)
}
