package service

import (
	"context"
	"errors"
	"fmt"
	"time"

	"github.com/nats-io/go-nats"
	nrpc "github.com/nats-rpc/gogo-nrpc"
	"xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/envelope"
)

var (
	// ErrDetach detach the process
	ErrDetach = errors.New("process is being detached")
)

// WithStartup is a service with a Startup() function
type WithStartup interface {
	Startup() error
}

// WithShutdown is a service with a Shutdown() function
type WithShutdown interface {
	Shutdown() error
}

// Inputs gives access to the input envelopes
type Inputs interface {
	List() []string
	Get(string) (envelope.Envelope, error)
}

// Outputs sends envelopes to the outputs
type Outputs interface {
	Forward(output string, id api.UUID) error
	Send(output string, envelope envelope.Envelope) error
}

type actorService struct {
	actor            *xbus.Actor
	workers          *nrpc.WorkerPool
	workersCtxCancel func()
	sub              *nats.Subscription
	surveySub        *nats.Subscription

	service interface{}
	process func(ctx context.Context, svc *actorService, req api.ActorProcessRequest, send func(api.ActorProcessingState)) error
}

func (svc actorService) handleSurvey(request api.SurveyRequest) {
	defer catchErrors(svc.actor.Client.Log)

	if request.OnlineStatus {
		if err := svc.actor.Client.API.Sentinel.ActorReady(svc.actor.ID.String()); err != nil {
			svc.actor.Client.Log.Errorf("Error calling 'ActorReady': %s", err)
		}
	}
}

func (svc *actorService) Process(
	ctx context.Context, req api.ActorProcessRequest, send func(api.ActorProcessingState),
) error {
	if svc.process == nil {
		return fmt.Errorf("process function undefined")
	}
	if err := svc.process(ctx, svc, req, send); err != nil {
		// Pass through a nrpc error
		if _, ok := err.(*nrpc.Error); ok {
			return err
		}
		send(api.ActorProcessingState{
			Context: *req.Context,
			Status:  api.ActorProcessingState_ERROR,
			Messages: []api.LogMessage{{
				Time:  time.Now(),
				Level: api.LogLevel_ERROR,
				Text:  err.Error(),
			}},
		})
	} else {
		send(api.ActorProcessingState{
			Context: *req.Context,
			Status:  api.ActorProcessingState_SUCCESS,
		})
	}
	return nil
}

func (svc *actorService) Startup() error {
	if i, ok := svc.service.(WithStartup); ok {
		if err := i.Startup(); err != nil {
			return err
		}
	}

	conn := svc.actor.Client.GetConn()

	ctx, cancel := context.WithCancel(context.Background())
	svc.workersCtxCancel = cancel
	workers := nrpc.NewWorkerPool(
		ctx, 1, 0, time.Second)
	svc.workers = workers

	handler := api.NewActorAgentConcurrentHandler(
		workers,
		conn,
		svc,
	)

	sub, err := conn.QueueSubscribe(
		api.XbusAPIPrefix+"actoragent."+svc.actor.ID.String()+".>",
		svc.actor.ID.String(),
		handler.Handler,
	)

	if err != nil {
		_ = svc.Shutdown()
		return err
	}
	svc.sub = sub

	if surveySub, err := svc.actor.Client.API.Sentinel.SurveySubscribe(
		svc.handleSurvey,
	); err == nil {
		svc.surveySub = surveySub
	} else {
		_ = svc.Shutdown()
		return err
	}

	if err := svc.actor.Client.API.Sentinel.ActorReady(svc.actor.ID.String()); err != nil {
		_ = svc.Shutdown()
		return err
	}

	return nil
}

func (svc *actorService) Shutdown() error {
	if svc.surveySub != nil {
		_ = svc.surveySub.Unsubscribe()
		svc.surveySub = nil
	}
	if svc.sub != nil {
		_ = svc.sub.Unsubscribe()
		svc.sub = nil
	}
	if svc.workers != nil {
		// cancel the workers context after a while if closing does not succeed
		// by itself
		t := time.AfterFunc(time.Second*2, func() {
			fmt.Println("Killing workers of ", svc.actor.Name)
			svc.workersCtxCancel()
		})
		svc.workers.Close(0)
		svc.workers = nil
		t.Stop()
	}
	if i, ok := svc.service.(WithShutdown); ok {
		if err := i.Shutdown(); err != nil {
			return err
		}
	}
	return nil
}
