package svctest

import (
	"context"
	"testing"

	"github.com/stretchr/testify/require"
	"orus.io/xbus/xbus-stdlib/service"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/testutils"
)

// WorkerTester helps testing a Worker
type WorkerTester struct {
	t        *testing.T
	factory  service.WorkerFactory
	actor    *api.Actor
	settings service.Settings
}

// Run calls a worker implementation with the given parameters, and
// returns the output envelopes
func (tester WorkerTester) Run(inputs Inputs) (*Outputs, error) {
	return tester.RunContext(context.Background(), inputs)
}

// RunContext calls a worker implementation with the given parameters, and
// returns the output envelopes
func (tester WorkerTester) RunContext(ctx context.Context, inputs Inputs) (*Outputs, error) {
	log := testutils.GetTestLogger(tester.t)
	outputs := NewOutputs()
	worker, err := tester.factory(tester.actor, tester.settings)
	if err != nil {
		return nil, err
	}
	if i, ok := worker.(service.WithStartup); ok {
		require.NoError(tester.t, i.Startup())
	}
	requestInputs := make([]*api.ActorProcessRequest_Input, 0, len(inputs))
	for name, input := range inputs {
		var (
			envelope *api.Envelope
			position *api.EnvelopePosition
		)
		if input != nil {
			envelope = &api.Envelope{
				ID: input.ID(),
			}
			p := api.EnvelopePositionStart(input.ID())
			position = &p
		}

		requestInput := &api.ActorProcessRequest_Input{
			Name:     name,
			Close:    input == nil,
			Envelope: envelope,
			Position: position,
		}

		requestInputs = append(requestInputs, requestInput)
	}
	err = worker.Process(ctx,
		api.ActorProcessRequest{
			Context: &api.ProcessingContext{
				ProcessID: DefaultProcessID,
				NodeID:    DefaultNodeID,
			},
			Inputs: requestInputs,
		},
		inputs, outputs, log.ZeroLogger)
	if i, ok := worker.(service.WithShutdown); ok {
		require.NoError(tester.t, i.Shutdown())
	}
	return outputs, err
}

// TestWorker returns a WorkerTester for a given WorkerFactory
func TestWorker(t *testing.T,
	factory service.WorkerFactory,
	actor *api.Actor, settings service.Settings) *WorkerTester {
	return &WorkerTester{
		t: t, factory: factory, actor: actor, settings: settings,
	}
}

// TestWorkerFunc returns a WorkerTester for a given WorkerFunc
func TestWorkerFunc(t *testing.T, fn service.WorkerFunc) *WorkerTester {
	return &WorkerTester{
		t: t, actor: nil, settings: nil,
	}
}
